@@ -1,6 +1,7 @@
 # Copyright (c) 2019 Alexander Todorov <atodorov@MrSenko.com>
 
 import os
+from datetime import datetime
 
 from . import TCMS
 
@@ -20,7 +21,7 @@ class Backend:  # pylint: disable=too-many-instance-attributes
 
             ... parse test results ...
 
-            test_case_id = backend.test_case_get_or_create(<description>)
+            test_case_id, _ = backend.test_case_get_or_create(<description>)
             backend.add_test_case_to_plan(test_case_id, backend.plan_id)
             test_execution_id = backend.add_test_case_to_run(test_case_id,
                                                              backend.run_id)
@@ -85,10 +86,36 @@ class Backend:  # pylint: disable=too-many-instance-attributes
             'name': 'CONFIRMED'
         })[0]['id']
 
+    def get_statuses_by_weight(self, lookup_condition):
+        """
+         Get a list of statuses based on lookup condition.
+
+         :param lookup_condition: ``tcms.testruns.models.TestExecutionStatus``
+            lookup condition
+         :type lookup_condition: dict
+         :rtype: list
+        """
+        return self.rpc.TestExecutionStatus.filter(lookup_condition)
+
+    def get_status_id_fallback(self, name):
+        """
+        Get status based on weight if name not found
+
+        :param name: ``tcms.testruns.models.TestExecutionStatus`` name
+        :type name: str
+        :rtype: int
+        """
+        if name in ['PASSED', 'WAIVED']:
+            lookup_condition = 'weight__gt'
+        elif name in ['FAILED', 'ERROR']:
+            lookup_condition = 'weight__lt'
+        return self.get_statuses_by_weight({lookup_condition: 0})[0]['id']
+
     def get_status_id(self, name):
         """
             Get the PK of ``tcms.testruns.models.TestExecutionStatus``
-            matching the test execution status.
+            matching the test execution status name or fallback based on
+            weight.
 
             .. important::
 
@@ -101,10 +128,12 @@ class Backend:  # pylint: disable=too-many-instance-attributes
             :rtype: int
         """
         if name not in self._statuses:
-            self._statuses[name] = self.rpc.TestExecutionStatus.filter({
-                'name': name
-            })[0]['id']
-
+            try:
+                self._statuses[name] = self.rpc.TestExecutionStatus.filter({
+                    'name': name
+                })[0]['id']
+            except IndexError:
+                self._statuses[name] = self.get_status_id_fallback(name)
         return self._statuses[name]
 
     def get_product_id(self, plan_id):
@@ -266,6 +295,30 @@ class Backend:  # pylint: disable=too-many-instance-attributes
 
         return plan_type[0]['id']
 
+    def external_plan_id(self):  # pylint: disable=no-self-use
+        """
+            Allows the user to specify `$TCMS_PLAN_ID` to point to an existing
+            TestPlan where new runs will be added!
+
+            .. warning::
+
+                Does not check if the specified TP exists!
+
+            :return: ``tcms.testplans.models.TestPlan`` PK or 0
+            :rtype: int
+        """
+        return os.environ.get('TCMS_PLAN_ID', 0)
+
+    def default_tester_id(self):
+        """
+            Used internally and by default this is the user sending the API
+            request. Plugins may want to override this.
+
+            :return: User ID
+            :rtype: int
+        """
+        return self.rpc.User.filter()[0]['id']
+
     def get_plan_id(self, run_id):
         """
             If a TestRun with PK `run_id` exists then return the TestPlan to
@@ -281,6 +334,10 @@ class Backend:  # pylint: disable=too-many-instance-attributes
             :return: ``tcms.testplans.models.TestPlan`` PK
             :rtype: int
         """
+        plan_id = self.external_plan_id()
+        if plan_id:
+            return plan_id
+
         result = self.rpc.TestRun.filter({'pk': run_id})
         if not result:
             product_id, product_name = self.get_product_id(0)
@@ -303,6 +360,7 @@ class Backend:  # pylint: disable=too-many-instance-attributes
                     'product_version': version_id,
                     'is_active': True,
                     'type': plan_type_id,
+                    'author': self.default_tester_id(),
                 })]
 
             # newly created TP
@@ -332,14 +390,19 @@ class Backend:  # pylint: disable=too-many-instance-attributes
             version_id, version_val = self.get_version_id(product_id)
             build_id, build_number = self.get_build_id(product_id, version_id)
             plan_id = self.get_plan_id(0)
-            # the user issuing the request
-            user_id = self.rpc.User.filter()[0]['id']
+            # TR.manager is always the author of the TP, which is either
+            # another existing user (existing TP) or self.default_tester_id()
+            # in case of newly created TP
+            manager_id = self.rpc.TestPlan.filter({
+                'pk': plan_id
+            })[0]['author_id']
 
             testrun = self.rpc.TestRun.create({
                 'summary': self.prefix + 'Results for %s, %s, %s' % (
                     product_name, version_val, build_number
                 ),
-                'manager': user_id,
+                'manager': manager_id,
+                'default_tester': self.default_tester_id(),
                 'plan': plan_id,
                 'build': build_id,
             })
@@ -352,6 +415,22 @@ class Backend:  # pylint: disable=too-many-instance-attributes
 
         return int(run_id)
 
+    def finish_test_run(self):
+        """
+            .. important::
+
+                Test runner plugins **may** call this method!
+
+            May be called at the end when there are no more test executions to
+            be sent to Kiwi TCMS. Default implementation will update
+            ``TR.stop_date``.
+
+            :return: None
+        """
+        self.rpc.TestRun.update(self.run_id, {
+            'stop_date': datetime.now().isoformat().replace('T', ' ')[:19],
+        })
+
     def test_case_get_or_create(self, summary):
         """
             Search for a TestCase with the specified `summary` and Product.
@@ -363,9 +442,11 @@ class Backend:  # pylint: disable=too-many-instance-attributes
 
             :param summary: A TestCase summary
             :type summary: str
-            :return: Serialized ``tcms.testcase.models.TestCase``
-            :rtype: dict
+            :return: Serialized ``tcms.testcase.models.TestCase`` and boolean
+                     flag to indicate if the TestCase has just been created!
+            :rtype: (dict, bool)
         """
+        created = False
         test_case = self.rpc.TestCase.filter({
             'summary': summary,
             'category__product': self.product_id,
@@ -380,8 +461,9 @@ class Backend:  # pylint: disable=too-many-instance-attributes
                 'notes': 'Created by tcms_api.plugin_helpers.Backend',
                 'is_automated': True,
             })]
+            created = True
 
-        return test_case[0]
+        return test_case[0], created
 
     def add_test_case_to_plan(self, case_id, plan_id):
         """
@@ -440,7 +522,10 @@ class Backend:  # pylint: disable=too-many-instance-attributes
             :type comment: str
             :return: None
         """
-        self.rpc.TestExecution.update(test_execution_id, {'status': status_id})
+        self.rpc.TestExecution.update(test_execution_id, {
+            'status': status_id,
+            'tested_by': self.default_tester_id(),
+        })
 
         if comment:
             self.add_comment(test_execution_id, comment)
